// Camera.cpp : Implementation of CCamera
#include "stdafx.h"
#include "KeckServer.h"
#include "Camera.h"
#include "imapi.h"
#include <fstream.h>

/////////////////////////////////////////////////////////////////////////////
// CCamera

#define HEIGHT 484
#define WIDTH  644
#define BANDS  1

#define MAX_CAMERAS 4
#define COLOR_CAM 3
#define MAX_BUFFERS 1000

#define FMT_TIF 0
#define FMT_PGM 1

long genesisDevice[MAX_CAMERAS];  
long genesisCamera[MAX_CAMERAS];  
long genesisThread[MAX_CAMERAS];
long ctrlBuf[MAX_CAMERAS];			// control buffer 
long grabOSB[MAX_CAMERAS][2];		// OSBs used for synchronization
long inBuf[MAX_CAMERAS][2];			// Double-buffered input buffer
long hostOutBuf[MAX_CAMERAS][MAX_BUFFERS];     // Double-buffered input buffer 
long numberOfHost[MAX_CAMERAS];

bool halt;

void InitCameras()
{
	// disable error catching
	imAppCatchError(IM_DEFAULT, NULL, NULL);

	// allocate the device,thread,and camera
	for (int c = 0; c < MAX_CAMERAS; c++) {
		imDevAlloc(c, 0, NULL, IM_DEFAULT, &genesisDevice[c]);
		imThrAlloc(genesisDevice[c], 0, &genesisThread[c]);
		imCamAlloc(genesisThread[c], "c:\\genesis\\dcf\\kodak\\es-310\\es310_60.dcf", IM_DEFAULT, &genesisCamera[c]);
	}

	// allocate input and control buffers in host memory 
	for (c = 0; c < MAX_CAMERAS; c++) {
		imBufAlloc(genesisThread[c], WIDTH, HEIGHT, BANDS, IM_UBYTE, IM_PROC, &inBuf[c][0]);
		imBufAlloc(genesisThread[c], WIDTH, HEIGHT, BANDS, IM_UBYTE, IM_PROC, &inBuf[c][1]);

		imBufAllocControl(genesisThread[c], &ctrlBuf[c]);

		numberOfHost[c] = 0;

		imSyncAlloc(genesisThread[c], &grabOSB[c][0]);
		imSyncAlloc(genesisThread[c], &grabOSB[c][1]);

		imBufPutField(genesisThread[c], ctrlBuf[c], IM_CTL_GRAB_MODE, IM_ASYNCHRONOUS);
	}
}

void CloseCameras()
{
	for (int c = 0; c < MAX_CAMERAS; c++)
	{
		for(int b = 0; b < numberOfHost[c]; b++)
			imBufFree(genesisThread[c], hostOutBuf[c][b]);
		imBufFree(genesisThread[c], inBuf[c][0]);
		imBufFree(genesisThread[c], inBuf[c][1]);
		imBufFree(genesisThread[c], ctrlBuf[c]);
		imSyncFree(genesisThread[c], grabOSB[c][0]);
		imSyncFree(genesisThread[c], grabOSB[c][1]);
		imCamFree(genesisThread[c], genesisCamera[c]);
		imThrFree(genesisThread[c]);
		imDevFree(genesisDevice[c]);
	}
}

STDMETHODIMP CCamera::SetDcfFile(/*[in, string]*/ unsigned char *dcf)
{
	for (int c = 0; c < MAX_CAMERAS; c++) {
		char full_dcf_file[1024];
		sprintf(full_dcf_file, "c:\\genesis\\dcf\\kodak\\es-310\\%s", dcf);
		imCamFree(genesisThread[c], genesisCamera[c]);
		imCamAlloc(genesisThread[c], (char *)full_dcf_file, IM_DEFAULT, &genesisCamera[c]);
	}

	return S_OK;
}

STDMETHODIMP CCamera::AllocBuffer(/*[in]*/ BOOL cameras[4], /*[in]*/ long numFrames)
{
	ReleaseBuffer();
	for (int c = 0; c < MAX_CAMERAS; c++) 
		if (cameras[c] && !numberOfHost[c])
		{
			numberOfHost[c] = numFrames;
			for (int i = 0; i < numberOfHost[c]; i++)
				imBufAlloc(genesisThread[c], WIDTH, HEIGHT, BANDS, IM_UBYTE, IM_HOST, &hostOutBuf[c][i]);
		}

	return S_OK;
}

STDMETHODIMP CCamera::ReleaseBuffer()
{
	for (int c = 0; c < MAX_CAMERAS; c++)
	{
		for(int b = 0; b < numberOfHost[c]; b++)
			imBufFree(genesisThread[c], hostOutBuf[c][b]);
		numberOfHost[c] = 0;
	}

	return S_OK;
}

STDMETHODIMP CCamera::StartCapture()
{
	imAppGetError(IM_ERR_RESET, 0);

	int maxHosts = 0;

	halt = false;
	
	// first grab
	for (int c = 0; c < MAX_CAMERAS; c++)
		if (numberOfHost[c])
		{
		    imDigGrab(genesisThread[c], 0, genesisCamera[c], inBuf[c][0], 1, ctrlBuf[c], grabOSB[c][0]);
			if (numberOfHost[c]>maxHosts)
				maxHosts = numberOfHost[c];
		}
	
    int b = 1;
	int hostIndex = 0;
    while (!halt && !imAppGetError(IM_ERR_CODE, NULL) && hostIndex < (maxHosts - 1)) {
		// queue next grab into other buffer
		for (c = 0; c < MAX_CAMERAS; c++)
			if (numberOfHost[c] - 1> hostIndex)
			    imDigGrab(genesisThread[c], 0, genesisCamera[c], inBuf[c][b], 1, ctrlBuf[c], grabOSB[c][b]);
    
		// switch buffers
        b = 1 - b;

		for (c = 0; c < MAX_CAMERAS; c++) 
			if (numberOfHost[c] - 1> hostIndex)
			{
				// process each frame as soon as the grab completes
				imSyncHost(genesisThread[c], grabOSB[c][b], IM_COMPLETED);

				// copy processed image to the host
				imBufCopy(genesisThread[c], inBuf[c][b], hostOutBuf[c][hostIndex], 0, 0);
		}
		hostIndex++;
    }

	// copy last images
	for (c = 0; c < MAX_CAMERAS; c++)
		if (numberOfHost[c] > hostIndex)
		{
			imSyncHost(genesisThread[c], grabOSB[c][1-b], IM_COMPLETED);
			imBufCopy(genesisThread[c], inBuf[c][1-b], hostOutBuf[c][hostIndex], 0, 0);
		}

	return S_OK;
}

STDMETHODIMP CCamera::SaveCapturedSeq(/*[in]*/ long format, /*[in, string]*/ unsigned char *baseFileName)
{
	char fileName[1024];
	unsigned char image[HEIGHT*WIDTH];
	FILE * out;
	for (int c = 0; c < MAX_CAMERAS; c++)
	{
		for(int b = 0; b < numberOfHost[c]; b++) 
		{
			switch(format)
			{
			case FMT_TIF:
				sprintf(fileName, "%s.%d.%04d.tif", baseFileName, c, b);
				imBufSave(genesisThread[c], fileName, IM_TIFF, hostOutBuf[c][b]);
				break;
			case FMT_PGM:
				sprintf(fileName, "%s.%d.%04d.pgm", baseFileName, c, b);
				imBufGet(genesisThread[c],hostOutBuf[c][b],image);
				out = fopen(fileName,"wb");
				fprintf(out,"P5\n%d %d 255\n",WIDTH,HEIGHT);
				fwrite(image,HEIGHT*WIDTH,1,out);
				fclose(out);
				break;
			}
		}
	}
	
	return S_OK;
}

STDMETHODIMP CCamera::GetCapturedImage(/*[in]*/ long cameraNumber, /*[in]*/ long imageNumber, /*[out]*/ unsigned char image[644*484])
{
	imBufGet(genesisThread[cameraNumber],hostOutBuf[cameraNumber][imageNumber],image);

	return S_OK;
}

STDMETHODIMP CCamera::GetLiveImage(/*[in]*/ long cameraNumber, /*[out]*/ unsigned char image[644*484])
{
	// grab an image into the buffer 0
    imDigGrab(genesisThread[cameraNumber], 0, genesisCamera[cameraNumber], inBuf[cameraNumber][0], 1, ctrlBuf[cameraNumber], grabOSB[cameraNumber][0]);
	imSyncHost(genesisThread[cameraNumber], grabOSB[cameraNumber][0], IM_COMPLETED);

	imBufGet(genesisThread[cameraNumber],inBuf[cameraNumber][0],image);

	return S_OK;
}

STDMETHODIMP CCamera::GetCamSettings(/*[in]*/ long cameraNumber, /*[in]*/ long item, /*[out]*/ double *value)
{
	imCamInquire(genesisThread[cameraNumber],genesisCamera[cameraNumber],item,value);
	return S_OK;
}

STDMETHODIMP CCamera::SetCamSettings(/*[in]*/ long cameraNumber, /*[in]*/ long item, /*[in]*/ double value)
{
	imCamControl(genesisThread[cameraNumber],genesisCamera[cameraNumber],item,value);
	return S_OK;
}

